/* This file is part of mailfrom filter.
   Copyright (C) 2006, 2007 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <setjmp.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <assert.h>
#include <fnmatch.h>
#include <syslog.h>
#include "mailfrom.h"
#include "prog.h"
#define DEFINE_BUILTIN_MODULE
#include "builtin.h"

#define SP(s) ((s) ? (s) : "''")


/* Code generation support */
#define CODE_INITIAL    128
#define CODE_INCREMENT  128

static prog_counter_t pc, pmax;
static instr_t *prog;
size_t stack_size = 4096;

size_t *string_off;        /* Offsets of string variables in var */
size_t string_count;       /* Number of string variables */

void
code_init()
{
	pmax = CODE_INITIAL;
	prog = malloc(pmax*sizeof prog[0]);
	if (!prog) {
		mu_error("Not enough memory to compile the program");
		exit(1);
	}
}

prog_counter_t
code_get_counter()
{
	return pc;
}

prog_counter_t
code_reserve(size_t count)
{
	prog_counter_t ret = pc;
	if (pc + count > pmax) {
		size_t incr = (pc + count - pmax + CODE_INCREMENT - 1)
 			       / CODE_INCREMENT;
		pmax += incr * CODE_INCREMENT;
		prog = realloc(prog, pmax*sizeof prog[0]);
		if (!prog) {
			mu_error("Not enough memory to compile the program");
			exit(1);
		}
	}
	pc += count;
	return ret;
}

prog_counter_t
code_instr(const instr_t ptr)
{
	if (pc >= pmax) {
		pmax += CODE_INCREMENT;
		prog = realloc(prog, pmax*sizeof prog[0]);
		if (!prog) {
			mu_error("Not enough memory to compile the program");
			exit(1);
		}
	}
	prog[pc] = ptr;
	return pc++;
}

prog_counter_t
code_immediate(const void *ptr)
{
	return code_instr(ptr);
}

void
code_put(prog_counter_t pos, void *ptr)
{
	assert(pos < pmax);
	prog[pos] = (instr_t) ptr;
}

instr_t
code_peek(prog_counter_t pos)
{
	assert(pos < pmax);
	return prog[pos];
}


unsigned long prog_trace_option = 0;
#define PROG_TRACE_ENGINE (prog_trace_option & BUILTIN_MASK_prog)

void
enable_module_trace(char *name)
{
	struct builtin_module *mod;
	
	for (mod = builtin_module; mod->name; mod++)
		if (strcmp(mod->name, name) == 0) {
			prog_trace_option |= mod->mask;
			return;
		}

	mu_error("No such module: %s", name);
}

void
disable_module_trace(char *name)
{
	struct builtin_module *mod;
	
	for (mod = builtin_module; mod->name; mod++)
		if (strcmp(mod->name, name) == 0) {
			prog_trace_option &= ~mod->mask;
			return;
		}

	mu_error("No such module: %s", name);
}

void
enable_prog_trace(const char *modlist)
{
	if (strcmp(modlist, "all") == 0) {
		prog_trace_option = ULONG_MAX;
		return;
	} else {
		char *modcopy, *s;
		modcopy = xstrdup(modlist);
	
		for (s = strtok(modcopy,","); s; s = strtok(NULL, ",")) 
			enable_module_trace(s);
		free(modcopy);
	}
}

void
disable_prog_trace(const char *modlist)
{
	if (strcmp(modlist, "all") == 0) {
		prog_trace_option = 0;
		return;
	} else {
		char *modcopy, *s;
		modcopy = xstrdup(modlist);
	
		for (s = strtok(modcopy,","); s; s = strtok(NULL, ",")) 
			disable_module_trace(s);
		free(modcopy);
	}
}

/* Run-time evaluation */

struct eval_environ {
	prog_counter_t pc;     /* Program counter */

	STKVAL *stack;         /* Stack space */
	prog_counter_t tos;    /* Top of stack: 0 <= tos <= toh */
	prog_counter_t toh;    /* Top of heap:  tos <= toh <= stack_size */
	prog_counter_t base;   /* Base pointer */
	STKVAL reg;            /* General purpose register */
	STKVAL *var;           /* Data space: %<name> variables */
	
	struct locus locus;    /* Program locus corresponding to PC */

	/* Sendmail interaction data: */
	SMFICTX *ctx;          /* Milter Context */
	void *data;            /* MTA symbol table */
	/* methods to access the latter */
	char *(*getsym)(void *data, char *str);
	int (*setreply)(void *data, char *code, char *xcode, char *message);
	void (*setheader)(void *data, struct old_header_node *hdr);

	/* Regular expression matching */
	regmatch_t *matches;   /* Match map */
	size_t matchcount;     /* Number of used entries in matches */
	size_t matchsize;      /* Total number of entries in matches */
	char *string;          /* Last matched string */

	mu_stream_t stream;    /* Capture stream */
	
	/* Non-local exits */
	prog_counter_t catch[mf_status_count];
	
	/* Built-in private data */
	void **bi_priv_array;
	
	/* Exit information */
	char *null_symbol;     /* Non resolved symbol that caused the abort */
	sfsistat status;       /* Program exit status */
	jmp_buf x;             /* Return point */
};

void
advance_pc(eval_environ_t env, unsigned cnt)
{
	env->pc += cnt;
}

void
adjust_stack(eval_environ_t env, unsigned cnt)
{
	env->tos -= cnt;
}

STKVAL *
env_reference(eval_environ_t env, storage_class_t sc, size_t n)
{
	switch (sc) {
	case storage_extern:
		return &env->var[n];
	case storage_auto:
		return &env->stack[env->base + n];
	case storage_param:
		return &env->stack[env->base - 2 - n];
	}
}

STKVAL *
env_var_ref(eval_environ_t env, size_t off)
{
	return env_reference(env, storage_extern, off);
}

void
env_var_inc(eval_environ_t env, size_t off)
{
	++*env_var_ref(env, off);
}       

const struct locus *
env_get_locus(eval_environ_t env)
{
	return &env->locus;
}

mu_stream_t
env_get_stream(eval_environ_t env)
{
	return env->stream;
}


void
prog_trace(eval_environ_t env, const char *fmt, ...)
{
	char buf[512];
	va_list ap;
	
	va_start(ap, fmt);
	vsnprintf(buf, sizeof buf, fmt, ap);
	va_end(ap);
	logmsg(LOG_DEBUG, "%4lu: %s:%lu: %s",
	       (unsigned long) env->pc,
	       env->locus.file, (unsigned long) env->locus.line,
	       buf);
}	

void
runtime_warning(eval_environ_t env, const char *fmt, ...)
{
	int n;
	va_list ap;
	char buf[512];
	
	n = snprintf(buf, sizeof buf, "RUNTIME WARNING near %s:%lu: ",
		     env->locus.file, (unsigned long) env->locus.line);
	va_start(ap, fmt);
	vsnprintf(buf + n, sizeof buf - n, fmt, ap);
	va_end(ap);
	mu_error(buf);
}
	      
void
runtime_error(eval_environ_t env, const char *fmt, ...)
{
	int n;
	va_list ap;
	char buf[512];
	
	n = snprintf(buf, sizeof buf, "RUNTIME ERROR near %s:%lu: ",
		     env->locus.file, (unsigned long) env->locus.line);
	va_start(ap, fmt);
	vsnprintf(buf + n, sizeof buf - n, fmt, ap);
	va_end(ap);
	mu_error(buf);
	env->status = SMFIS_TEMPFAIL; /* FIXME */
	longjmp(env->x, 1);
}
	      
void *
get_immediate(eval_environ_t env, unsigned n)
{
	return prog[env->pc + n + 1];
}

STKVAL
get_arg(eval_environ_t env, unsigned n)
{
	return env->stack[env->tos - n - 1];
}

void
push(eval_environ_t env, STKVAL val)
{
	if (env->tos >= env->toh) {
		debug2(100, "tos=%lu, toh=%lu",
		       (unsigned long) env->tos,
		       (unsigned long) env->toh);
		runtime_error(env,
			      "Out of stack space; increase #pragma stacksize");
	}
	env->stack[env->tos++] = val;
}

STKVAL
pop(eval_environ_t env)
{
	if (env->tos == 0)
		runtime_error(env, "Stack underflow");
	return env->stack[--env->tos];
}

#define B2STACK(s) (s + sizeof(STKVAL) - 1) / sizeof(STKVAL)

STKVAL
heap_reserve(eval_environ_t env, size_t size)
{
	size_t words = B2STACK(size);
	
        if (env->toh - words <= env->tos) {
                /* Heap overrun: */
                /* gc(); */
                if (env->tos - words <= env->tos) 
                        runtime_error(env,
				      "Heap overrun; increase #pragma stacksize");
        }
        env->toh -= words;
        return env->stack + env->toh--;
}

STKVAL
heap_tempspace(eval_environ_t env, size_t size)
{
	size_t words = B2STACK(size);
	
        if (env->toh - words <= env->tos) {
                /* Heap overrun: */
                /* gc(); */
                if (env->tos - words <= env->tos) 
                        runtime_error(env,
				      "Heap overrun; increase #pragma stacksize");
        }
        return env->stack + env->toh - words;
}


STKVAL
heap_max_tempspace(eval_environ_t env, size_t *size)
{
	*size = (env->toh - env->tos - 1) * sizeof env->stack[0];
	return env->stack + env->tos + 1;
}

STKVAL
heap_fix_tempspace(eval_environ_t env, size_t size)
{
	size_t words = B2STACK(size);
	size_t toh;
	if (words >= env->toh - env->tos)
		runtime_error(env,
			      "INTERNAL ERROR: memory chunk too big "
			      "to fit into heap");
	env->toh -= words;
	memmove(env->stack + env->toh, env->stack + env->tos + 1, size);
	return env->stack + env->toh--;
}

void
pushs(eval_environ_t env, char *s)
{
	push(env, (STKVAL) strcpy(heap_reserve(env, strlen(s) + 1), s));
}

/* Auxiliary instructions */
void
instr_locus(eval_environ_t env)
{
	env->locus.file = get_immediate(env, 0);
	env->locus.line = (size_t) get_immediate(env, 1);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "LOCUS");
	advance_pc(env, 2);
}

void
dump_locus(prog_counter_t i, prog_counter_t *ret __attribute__((__unused__)))
{
	printf("\"%s\" %lu", (char*) prog[i],
	       (unsigned long) (size_t) prog[i+1]);
}

/* Stack manipulation instructions */
void
instr_xchg(eval_environ_t env)
{
	prog_counter_t p = env->tos - 1;
	STKVAL tmp = env->stack[p];
	env->stack[p] = env->stack[p - 1];
	env->stack[p - 1] = tmp;
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "XCHG");
}

void
instr_dup(eval_environ_t env)
{
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "DUP");
	push(env, get_arg(env, 0));
}

void
instr_pop(eval_environ_t env)
{
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "POP");
	pop(env);
}

void
instr_push_immediate(eval_environ_t env)
{
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "PUSH_IMMEDIATE %p", get_immediate(env, 0));
	push(env, get_immediate(env, 0));
	advance_pc(env, 1);
}

void
dump_push_immediate(prog_counter_t i,
		    prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%p", prog[i]); /* FIXME: Bad. Should have a way for determining
				  operand type */
}

void
instr_stkalloc(eval_environ_t env)
{
	unsigned n = (unsigned) get_immediate(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "STKALLOC %p", n);
	env->tos += n;
	advance_pc(env, 1);
}

void
dump_stkalloc(prog_counter_t i,
	      prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%u", (unsigned)prog[i]);
}

void
instr_backref(eval_environ_t env)
{
	unsigned n = (unsigned) get_immediate(env, 0);
	char *s;
	size_t matchlen;
	
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "BACKREF %u", (unsigned)get_immediate(env, 0));
	advance_pc(env, 1);
	
	if (!env->matches || !env->string) {
		/* FIXME: Try to throw exception first:
		   if (env_catch(env, mf_no_regex) == 0)
			return 0;
		*/
		runtime_error(env, "No previous regular expression");
	}
	if (n > env->matchcount || env->matches[n].rm_so == -1) {
		/* FIXME: See above */
		runtime_error(env, "Invalid back-reference number");
	}

	matchlen = env->matches[n].rm_eo - env->matches[n].rm_so;
	s = heap_reserve(env, matchlen + 1);
	memcpy(s, env->string + env->matches[n].rm_so, matchlen);
	s[matchlen] = 0;
	push(env, (STKVAL) s);
}

void
dump_backref(prog_counter_t i,
	     prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%u", (unsigned) prog[i]);
}

/* Type conversion instructions */
void
instr_ston(eval_environ_t env)
{
	char *s = pop(env);
	char *p;
	long v = strtol(s, &p, 0);

	if (PROG_TRACE_ENGINE)
		prog_trace(env, "STON %s", s);
	if (*p) {
		if (env_catch(env, mf_ston_conv) == 0)
			return;
		runtime_error(env,
		    "Cannot convert stack value to number (stopped at %8.8s)",
			      p);
	}
	push(env, (STKVAL) v);
}

void
instr_ntos(eval_environ_t env)
{
	long v = (long) get_arg(env, 0);
	char buf[NUMERIC_BUFSIZE_BOUND];
	
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "NTOS %ld", v);
	adjust_stack(env, 1);
	
	snprintf(buf, sizeof buf, "%ld", v);
	pushs(env, buf);
}

/* Evaluation instructions */
void
instr_cmp(eval_environ_t env)
{
	long l = (long) get_arg(env, 1);
	long r = (long) get_arg(env, 0);
	adjust_stack(env, 2);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "CMP %ld %ld", l, r);
	push(env, (STKVAL) (l == r));
}

void
instr_symbol(eval_environ_t env)
{
	char *symbol = get_immediate(env, 0);
	char *s = env->getsym(env->data, symbol);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "SYMBOL %s", symbol);
	if (!s) {
		if (PROG_TRACE_ENGINE)
			prog_trace(env, "%s is not defined", symbol);
		env->null_symbol = symbol;
		longjmp(env->x, 1);
	} else if (env->null_symbol && strcmp(env->null_symbol, symbol) == 0) 
		env->null_symbol = NULL;
	
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "%s dereferenced to %s", symbol, s);
		
	pushs(env, (STKVAL) s);
	advance_pc(env, 1);
}

void
dump_symbol(prog_counter_t i, prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%s", (char *) prog[i]);
}

/* Comparation instructions */
void
instr_eqn(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "EQN %ld %ld", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (a == b));
}

void
instr_eqs(eval_environ_t env)
{
	char *a = (char*) get_arg(env, 1);
	char *b = (char*) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "EQS %s %s", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (strcmp(a, b) == 0));
}

void
instr_nen(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "NEN %ld %ld", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (a != b));
}
	
void
instr_nes(eval_environ_t env)
{
	char *a = (char*) get_arg(env, 1);
	char *b = (char*) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "NES %s %s", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) strcmp(a, b));
}

void
instr_ltn(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "LTN %ld %ld", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (a < b));
}

void
instr_lts(eval_environ_t env)
{
	char *a = (char*) get_arg(env, 1);
	char *b = (char*) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "LTS %s %s", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (strcmp(a, b) < 0));
}

void
instr_len(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "LEN %ld %ld", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (a <= b));
}

void
instr_les(eval_environ_t env)
{
	char *a = (char*) get_arg(env, 1);
	char *b = (char*) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "LES %s %s", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (strcmp(a, b) <= 0));
}

void
instr_gtn(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "GTN %ld %ld", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (a > b));
}

void
instr_gts(eval_environ_t env)
{
	char *a = (char*) get_arg(env, 1);
	char *b = (char*) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "GTS %s %s", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (strcmp(a, b) > 0));
}

void
instr_gen(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "GEN %ld %ld", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (a >= b));
}

void
instr_ges(eval_environ_t env)
{
	char *a = (char*) get_arg(env, 1);
	char *b = (char*) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "GES %s %s", a, b); 
	adjust_stack(env, 2);
	push(env, (STKVAL) (strcmp(a, b) >= 0));
}

/* Jump instructions */
void
instr_bz(eval_environ_t env)
{
	long v = (long) get_arg(env, 0);
	long off = (long) get_immediate(env, 0);

	if (PROG_TRACE_ENGINE)
		prog_trace(env, "BZ %ld (%ld)", off, v);
	adjust_stack(env, 1);
	if (v == 0)
		advance_pc(env, off);
	advance_pc(env, 1);
}

void
dump_branch (prog_counter_t i, prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%ld (%ld)", (long) prog[i], i + (long) prog[i] + 1);
}

void
instr_bnz(eval_environ_t env)
{
	long v = (long) get_arg(env, 0);
	long off = (long) get_immediate(env, 0);

	if (PROG_TRACE_ENGINE)
		prog_trace(env, "BNZ %ld (%ld)", off, v);
	adjust_stack(env, 1);
	if (v != 0)
		advance_pc(env, off);
	advance_pc(env, 1);
}

void
instr_jmp(eval_environ_t env)
{
	long off = (long) get_immediate(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "JMP %ld", off);
	advance_pc(env, off+1);
}

/* Boolean instructions */
void
instr_not(eval_environ_t env)
{
	long v = (long) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "NOT %ld", v);
	adjust_stack(env, 1);
	push(env, (STKVAL) !v);
}

/* Bitwise arithmetics */
void
instr_logand(eval_environ_t env)
{
	unsigned long a = (unsigned long) get_arg(env, 1);
	unsigned long b = (unsigned long) get_arg(env, 0);
	adjust_stack(env, 2);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "LOGAND %lu %lu", a, b); 
	push(env, (STKVAL) (a & b));
}

void
instr_logor(eval_environ_t env)
{
	unsigned long a = (unsigned long) get_arg(env, 1);
	unsigned long b = (unsigned long) get_arg(env, 0);
	adjust_stack(env, 2);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "LOGOR %lu %lu", a, b); 
	push(env, (STKVAL) (a | b));
}

void
instr_logxor(eval_environ_t env)
{
	unsigned long a = (unsigned long) get_arg(env, 1);
	unsigned long b = (unsigned long) get_arg(env, 0);
	adjust_stack(env, 2);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "LOGXOR %lu %lu", a, b); 
	push(env, (STKVAL) (a ^ b));
}

void
instr_lognot(eval_environ_t env)
{
	unsigned long v = (unsigned long) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "LOGNOT %ld", v);
	adjust_stack(env, 1);
	push(env, (STKVAL) ~v);
}

/* Arithmetic instructions */
void
instr_add(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	adjust_stack(env, 2);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "ADD %ld %ld", a, b); 
	push(env, (STKVAL) (a + b));
}

void
instr_sub(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	adjust_stack(env, 2);
	if (PROG_TRACE_ENGINE)
                prog_trace(env, "SUB %ld %ld", a, b); 
	push(env, (STKVAL) (a - b));
}

void
instr_mul(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	adjust_stack(env, 2);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "MUL %ld %ld", a, b); 
	push(env, (STKVAL) (a * b));
}

void
instr_div(eval_environ_t env)
{
	long a = (long) get_arg(env, 1);
	long b = (long) get_arg(env, 0);
	adjust_stack(env, 2);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "DIV %ld %ld", a, b);
	if (b == 0) {
		if (env_catch(env, mf_divzero) == 0)
			return;
		runtime_error(env, "Division by zero");
	}
	push(env, (STKVAL) (a / b));
}

void
instr_neg(eval_environ_t env)
{
	long v = (long) get_arg(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "NEG %ld", v);
	adjust_stack(env, 1);
	push(env, (STKVAL) -v);
}

/* Matching and Regular expression instructions */
void
instr_regex(eval_environ_t env)
{
	int v;
	long flag = (long)get_arg(env, 0);
	regex_t *re = get_arg(env, 1);
	char *string = get_arg(env, 2);
	adjust_stack(env, 3);
	
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "REGEX %p %s", re, string);

	env->matchcount = re->re_nsub;
	if (env->matchsize < env->matchcount + 1) {
		void *p = realloc(env->matches,
				  sizeof(env->matches[0])
				  * (env->matchcount + 1));
		if (!p)
			runtime_error(env, "Not enough memory");
		env->matches = p;
		env->matchsize = env->matchcount + 1;
	}
	
	v = regexec(re, string, env->matchcount + 1, env->matches, 0);
	env->string = string;
	
	if (flag) {
		regfree(re);
		free(re);
	}
	push(env, (STKVAL) (v == 0));
}

void
instr_regcomp(eval_environ_t env)
{
	int v;
	regex_t *re;
	char *expr = get_arg(env, 0);
	adjust_stack(env, 1);

	if (PROG_TRACE_ENGINE)
		prog_trace(env, "REGCOMP %s", expr);
	re = malloc(sizeof *re);
	if (!re)
		runtime_error(env, "Not enough memory");
	v = regcomp(re, expr, regex_flags);
	if (v) {
		char errbuf[512];
		regerror(v, re, errbuf, sizeof(errbuf));
		free(re);
		if (env_catch(env, mf_regcomp) == 0)
			return;
		runtime_error(env, "compiling regex `%s': %s",
			      expr,
			      errbuf);
	}	
	push(env, (STKVAL) re);
}

void
instr_fnmatch(eval_environ_t env)
{
	char *string = get_arg(env, 1);
	char *pattern = get_arg(env, 0);
	adjust_stack(env, 2);
	if (PROG_TRACE_ENGINE)
                prog_trace(env, "FNMATCH %s %s", string, pattern);
	push(env, (STKVAL) (fnmatch (pattern, string, 0) == 0));
}

static int
mx_match(eval_environ_t env, char *string,
	 int (*matcher)(const char *name, void *data), void *data)
{
	int rc = 0;
	mxbuf_t mxbuf;
	mf_status mxstat;

	char *p = strchr(string, '@');
	if (p)
		p++;
	else
		p = string;
	mxstat = getmx(p, mxbuf);
	rc = 0;
	if (mxstat == mf_success) {
		int i;
		
		for (i = 0; i < MAXMXCOUNT && mxbuf[i]; i++) {
			if (matcher(mxbuf[i], data)) {
				rc = 1;
				break;
			}
		}
		freemx(mxbuf);
	} else
		env_catch(env, mxstat);
	return rc;
}

static int
fn_matcher(const char *string, void *data)
{
	return fnmatch (data, string, 0) == 0;
}

void
instr_fnmatch_mx(eval_environ_t env)
{
	char *string = get_arg(env, 1);
	char *pattern = get_arg(env, 0);
	adjust_stack(env, 2);

	if (PROG_TRACE_ENGINE)
		prog_trace(env, "FNMATCH,MX %s %s", string, pattern);
	push(env, (STKVAL) mx_match(env, string, fn_matcher, pattern));
}

static int
regex_matcher(const char *string, void *data)
{
	return regexec(data, string, 0, NULL, 0) == 0;
}

void
instr_regex_mx(eval_environ_t env)
{
	int rc;
	long flag = (long)get_arg(env, 0);
	regex_t *re = get_arg(env, 1);
	char *string = get_arg(env, 2);
	adjust_stack(env, 3);
	
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "REGEX,MX %p %s", re, string);
	
	rc = mx_match(env, string, regex_matcher, re);
	
	if (flag) {
		regfree(re);
		free(re);
	}
	push(env, (STKVAL) rc);
}
	
/* Mail filter specific instructions */

void
instr_next(eval_environ_t env)
{
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "NEXT", env->locus.file, env->locus.line);
	trace("%s%s:%lu: next",
	      mailfromd_msgid(env->ctx),
	      env->locus.file, env->locus.line);
}

void
instr_result(eval_environ_t env)
{
	sfsistat status = (sfsistat) get_immediate(env, 0);
	char *code = (char *) get_immediate(env, 1);
	char *xcode = (char *) get_immediate(env, 2);
	char *message = (char *) get_arg(env, 0);

	if (PROG_TRACE_ENGINE)
		prog_trace(env, "RESULT %d %s %s %s",
			   status,
			   SP(code),
			   SP(xcode),
			   SP(message));
	
	trace("%s%s:%lu: %s %s %s %s",
	      mailfromd_msgid(env->ctx),
	      env->locus.file, env->locus.line,
	      sfsistat_str(status),
	      SP(code),
	      SP(xcode),
	      SP(message));
	
	env->status = status;
	env->setreply(env->data, code, xcode, message);
	advance_pc(env, 3);
	adjust_stack(env, 1);
}

void
dump_result(prog_counter_t i, prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%s %s %s",
	       sfsistat_str((sfsistat)prog[i]),
	       SP((char*) prog[i+1]),
	       SP((char*) prog[i+2]));
}

void
instr_header(eval_environ_t env)
{
	struct old_header_node *hdr = xmalloc (sizeof(*hdr));

	hdr->opcode = (enum header_opcode) get_immediate(env, 0);
	hdr->name = (char *) get_immediate(env, 1);
	hdr->value = strdup((char *) get_arg(env, 0));
	
	advance_pc(env, 2);
	adjust_stack(env, 1);
	
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "HEADER %d %s %s",
			   hdr->opcode, hdr->name, hdr->value);
	
	trace("%s%s:%lu: %s %s %s",
	      mailfromd_msgid(env->ctx),
	      env->locus.file, env->locus.line,
	      header_command_str(hdr->opcode),
	      hdr->name, hdr->value);
	
	env->setheader(env->data, hdr);
}

void
dump_header(prog_counter_t i, prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%s %s %s",
	       header_command_str((enum header_opcode) prog[i]),
	       SP((char*) prog[i+1]), SP((char*) prog[i+2]));
}

void
instr_builtin(eval_environ_t env)
{
	char *name = get_immediate(env, 0);
	void (*handler)(eval_environ_t) = get_immediate(env, 1);

	if (PROG_TRACE_ENGINE)
		prog_trace(env, "BUILTIN %s", name);
	advance_pc(env, 2);
	handler(env);
}

void
dump_builtin(prog_counter_t i, prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%s ", SP((char*) prog[i]));
}

void
instr_concat(eval_environ_t env)
{
	char *left = get_arg(env, 1);
	char *right = get_arg(env, 0);
	char *res = heap_reserve(env, strlen(left) + strlen(right) + 1);

	strcat(strcpy(res, left), right);
	adjust_stack(env, 2);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "CONCAT ('%s','%s')='%s'", left, right, res);
	push(env, (STKVAL) res);
}

void
instr_variable(eval_environ_t env)
{
	size_t off = (size_t) get_immediate(env, 0);
	advance_pc(env, 1);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "VARIABLE %lu", (unsigned long) env->var[off]);
	push(env, (STKVAL) env->var[off]);
}

void
dump_variable(prog_counter_t i,
	      prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%lu ", (unsigned long) prog[i]);
}

void
instr_asgn(eval_environ_t env)
{
	STKVAL val = get_arg(env, 0);
	storage_class_t sc = (storage_class_t) get_immediate(env, 0);
	size_t off = (size_t) get_immediate(env, 1);
	advance_pc(env, 2);
	adjust_stack(env, 1);
	if (PROG_TRACE_ENGINE) 
		prog_trace(env, "ASGN %s %lu %lu",
			   storage_class_str(sc),
			   (unsigned long) off, (unsigned long) val);
	*env_reference(env, sc, off) = val;
}

void
dump_asgn(prog_counter_t i,
	  prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%s %lu", storage_class_str((storage_class_t)prog[i]),
	       (unsigned long) prog[i+1]);
}

void
instr_svalidate(eval_environ_t env)
{
	STKVAL val = get_arg(env, 0);
	if (prog_trace_option)
		prog_trace(env, "SVALIDATE %p", val);
	if (val == 0) {
		static char nullstr[] = "";
		env->stack[env->tos - 1] = (STKVAL)nullstr;
	}
}		

void
instr_arg(eval_environ_t env)
{
	size_t n = (size_t) get_immediate(env, 0);
	STKVAL val = *env_reference(env, storage_param, n);
	advance_pc(env, 1);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "ARG %lu = %p", 
			   (unsigned long) n, val);
	push(env, val);
}

void
instr_auto(eval_environ_t env)
{
	size_t n = (size_t) get_immediate(env, 0);
	STKVAL val = *env_reference(env, storage_auto, n);
	advance_pc(env, 1);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "AUTO %lu = %p", 
			   (unsigned long) n, val);
	push(env, val);
}

void
instr_catch(eval_environ_t env)
{
	long off = (long) get_immediate(env, 0);
	/* int count = (long) get_immediate(env, 1); -- Count is not used */
	unsigned mask = (unsigned) get_immediate(env, 2);
	int i;
	prog_counter_t entry = env->pc + 4;
	
	if (PROG_TRACE_ENGINE) 
		prog_trace(env, "CATCH %ld, %ld", entry, off);

	for (i = 0; i < mf_status_count; i++)
		if (mask & EXMASK(i)) {
			if (PROG_TRACE_ENGINE)
				prog_trace(env, "CATCH TARGET: %s",
					   mf_status_str((mf_status)i));
			env->catch[i] = entry;
		}

	advance_pc(env, off);
}

void
dump_catch(prog_counter_t i, prog_counter_t *reti)
{
	unsigned mask = (unsigned) prog[i+2];
	printf("%ld (%ld)", (long) prog[i], i + (long) prog[i]);
	printf("; Targets: ");

	for (i = 0; i < mf_status_count; i++)
		if (mask & EXMASK(i)) 
			printf("%s ", mf_status_str((mf_status)i));
}

void
instr_echo(eval_environ_t env)
{
	char *str = (char*) get_arg(env, 0);
	logmsg(LOG_INFO, "%s", str);	
}

void
instr_return(eval_environ_t env)
{
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "RETURN");
	env_leave_frame(env, 0);
	env->pc--;
}

void
instr_retcatch(eval_environ_t env)
{
	prog_counter_t pc = env->pc;
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "RETCATCH");
	env_leave_frame(env, 1);
	env->pc = pc;
}

void
instr_saveex(eval_environ_t env)
{
	unsigned exmask = (unsigned) get_immediate(env, 0);
	int i;
	advance_pc(env, 1);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "SAVEEX %x", exmask);
	push(env, (STKVAL) exmask);
	for (i = 0; i < mf_status_count; i++) {
		if (EXMASK(i) & exmask) 
			push(env, (STKVAL) env->catch[i]);
	}
}

static void
dump_saveex(prog_counter_t i, prog_counter_t *ret)
{
	printf("%x", (unsigned)prog[i]);
}

void
instr_restex(eval_environ_t env)
{
	unsigned exmask = (unsigned)env->stack[env->base];
	int i;
	for (i = 0; i < mf_status_count; i++) {
		if (EXMASK(i) & exmask)
			env->catch[i] = (prog_counter_t)
				           env->stack[env->base+i+1];
	}
}

void
instr_adjust(eval_environ_t env)
{
	long nargs = (long) get_immediate(env, 0);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "ADJUST %ld", nargs);
	adjust_stack(env, nargs);
	advance_pc(env, 1);
}

void
instr_popreg(eval_environ_t env)
{
	env->reg = pop(env);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "POPREG %p", env->reg);
}

void
instr_pushreg(eval_environ_t env)
{
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "PUSHREG %p", env->reg);
	push(env, env->reg);
}

void
instr_funcall(eval_environ_t env)
{
	char *name = get_immediate(env, 0);
	prog_counter_t pc = (prog_counter_t) get_immediate(env, 1);
	advance_pc(env, 2);
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "FUNCALL %s (%lu)", name, (unsigned long)pc);
	env_make_frame(env);
	env->pc = pc-1;
}

void
dump_funcall(prog_counter_t i, prog_counter_t *ret __attribute__((__unused__)))
{
	printf("%s (%lu)", SP((char*) prog[i]), (unsigned long)prog[i+1]);
}

/* Opcode:

   xlat N x y x1 y1 ... xN yN

   Synopsis:

   Scan the table until xI==REG is found or the table is exhausted.
   If found, replace REG with yI and return 0. Otherwise, return 1
*/
void
instr_xlat(eval_environ_t env)
{
	long i, count = (long) get_immediate(env, 0);
	instr_t *tab = prog + env->pc + 2;
	count *= 2;
	advance_pc(env, count+1);
        if (PROG_TRACE_ENGINE)
		prog_trace(env, "XLAT %ld", count);
	for (i = 0; i < count; i += 2) {
		if ((long) tab[i] == (long) env->reg) {
			env->reg = (STKVAL) tab[i+1];
			push(env, (STKVAL)0);
			return;
		}
	}
	push(env, (STKVAL)1);
}

void
instr_xlats(eval_environ_t env)
{
	long i, count = (long) get_immediate(env, 0);
	instr_t *tab = prog + env->pc + 2;
	char *str = (char*) env->reg;
	count *= 2;
	advance_pc(env, count+1);
        if (PROG_TRACE_ENGINE)
		prog_trace(env, "XLATS %ld", count);
	for (i = 0; i < count; i += 2) {
		if (strcmp((char*) tab[i], str) == 0) {
			env->reg = (STKVAL) tab[i+1];
			push(env, (STKVAL)0);
			return;
		}
	}
	push(env, (STKVAL)1);
}

void
dump_xlat(prog_counter_t i, prog_counter_t *ret)
{
	long j, count = (long) prog[i];
	count *= 2;
	*ret = i + count;
	printf("%ld ", (long) prog[i]);
	i++;
	for (j = 0; j < count; j += 2) 
		printf("(%ld %ld) ", (long)prog[i+j], (long)prog[i+j+1]);
}

void
dump_xlats(prog_counter_t i, prog_counter_t *ret)
{
	long j, count = (long) prog[i];
	count *= 2;
	*ret = i + count;
	printf("%ld ", (long) prog[i]);
	i++;
	for (j = 0; j < count; j += 2) 
		printf("(%s %ld) ", (char*)prog[i+j], (long)prog[i+j+1]);
}

void
instr_jreg(eval_environ_t env)
{
	env->pc += (prog_counter_t)env->reg - 1;
	if (PROG_TRACE_ENGINE)
		prog_trace(env, "JREG %ld (%lu)",
			   (long)((prog_counter_t)env->reg - 1),
			   (unsigned long)env->pc);
}


struct optab {
	char *name;
	instr_t instr;
	void (*dump) (prog_counter_t i, prog_counter_t *p);
	long length;
} optab[] = {
	{ "XCHG",           instr_xchg,           NULL,                  0 },
        { "STON",           instr_ston,           NULL,                  0 },
        { "NTOS",           instr_ntos,           NULL,                  0 },
        { "LOCUS",          instr_locus,          dump_locus,            2 },
        { "PUSH_IMMEDIATE", instr_push_immediate, dump_push_immediate,   1 },
	{ "STKALLOC",       instr_stkalloc,       dump_stkalloc,         1 },
        { "SYMBOL",         instr_symbol,         dump_symbol,           1 },
        { "BZ",             instr_bz,             dump_branch,           1 },
        { "BNZ",            instr_bnz,            dump_branch,           1 },
        { "JMP",            instr_jmp,            dump_branch,           1 },
        { "DUP",            instr_dup,            NULL,                  0 },
        { "CMP",            instr_cmp,            NULL,                  0 },
        { "POP",            instr_pop,            NULL,                  0 },
        { "REGEX",          instr_regex,          NULL,                  0 },
        { "REGEX,MX",       instr_regex_mx,       NULL,                  0 },
        { "REGCOMP",        instr_regcomp,        NULL,                  0 },
        { "FNMATCH",        instr_fnmatch,        NULL,                  0 },
        { "FNMATCH,MX",     instr_fnmatch_mx,     NULL,                  0 },
        { "NOT",            instr_not,            NULL,                  0 },
        { "NEG",            instr_neg,            NULL,                  0 },
        { "NEXT",           instr_next,           NULL,                  0 },
        { "RESULT",         instr_result,         dump_result,           3 },
        { "HEADER",         instr_header,         dump_header,           2 },
        { "EQN",            instr_eqn,            NULL,                  0 },
        { "EQS",            instr_eqs,            NULL,                  0 },
        { "NEN",            instr_nen,            NULL,                  0 },
        { "NES",            instr_nes,            NULL,                  0 },
	{ "LTN",            instr_ltn,            NULL,                  0 },
	{ "LTS",            instr_lts,            NULL,                  0 },
	{ "LEN",            instr_len,            NULL,                  0 },
	{ "LES",            instr_les,            NULL,                  0 },
	{ "GTN",            instr_gtn,            NULL,                  0 },
	{ "GTS",            instr_gts,            NULL,                  0 },
	{ "GEN",            instr_gen,            NULL,                  0 },
	{ "GES",            instr_ges,            NULL,                  0 },
	{ "ADD",            instr_add,            NULL,                  0 },
	{ "SUB",            instr_sub,            NULL,                  0 },
	{ "MUL",            instr_mul,            NULL,                  0 },
	{ "DIV",            instr_div,            NULL,                  0 },
	{ "BUILTIN",        instr_builtin,        dump_builtin,          2 },
	{ "CONCAT",         instr_concat,         NULL,                  0 },
	{ "SET",            instr_asgn,           dump_asgn,             2 },
	{ "VARIABLE",       instr_variable,       dump_variable,         1 },
	{ "SVALIDATE",      instr_svalidate,      NULL,                  0 },
	{ "ARG",            instr_arg,            dump_variable,         1 },
	{ "AUTO",           instr_auto,           dump_variable,         1 },
	{ "CATCH",          instr_catch,          dump_catch,            3 },
	{ "ECHO",           instr_echo,           NULL,                  0 },
	{ "RETURN",         instr_return,         NULL,                  0 },
	{ "RETCATCH",       instr_retcatch,       NULL,                  0 }, 
	{ "ADJUST",         instr_adjust,         dump_variable,         1 },
	{ "POPREG",         instr_popreg,         NULL,                  0 },
	{ "PUSHREG",        instr_pushreg,        NULL,                  0 },
	{ "FUNCALL",        instr_funcall,        dump_funcall,          2 },
	{ "JREG",           instr_jreg,           NULL,                  0 },
	{ "XLAT",           instr_xlat,           dump_xlat,             0 },
	{ "XLATS",          instr_xlats,          dump_xlats,            0 },
	{ "SAVEEX",         instr_saveex,         dump_saveex,           1 },
	{ "RESTEX",         instr_restex,         NULL,                  0 },
	{ "BACKREF",        instr_backref,        dump_backref,          1 },
	{ "LOGAND",         instr_logand,         NULL,                  0 },
	{ "LOGOR",          instr_logor,          NULL,                  0 },
	{ "LOGXOR",         instr_logxor,         NULL,                  0 },
	{ "LOGNOT",         instr_lognot,         NULL,                  0 },
	{ NULL }
};

struct optab *
find_op(instr_t instr)
{
	struct optab *op;
	for (op = optab; op->instr; op++)
		if (op->instr == instr)
			return op;
	return NULL;
}

void
dump_code(prog_counter_t start, prog_counter_t end)
{
	prog_counter_t i;
	if (end == 0)
		end = pc;
	for (i = start; i < end; i++) {
		printf("%4lu: ", i);
		if (prog[i]) {
			struct optab *op = find_op (prog[i]);
			if (!op) {
				printf("Unknown function. Aborting dump\n");
				exit (1);
			}
			printf("%s ", op->name);
			if (op->dump)
				op->dump (i+1, &i);
			i += op->length;
		} else
			printf("STOP");
		putchar('\n');
	}
}


struct declvar {
	struct declvar *next;
	char *name;
	size_t off;
	STKVAL value;
};

static struct declvar *declvar;

int
variable_declared_p(const char *name)
{
	struct declvar *dv;
	for (dv = declvar; dv; dv = dv->next)
		if (strcmp(dv->name, name) == 0)
			return 1;
	return 0;
}

void
declare_variable(char *name, size_t off, STKVAL value)
{
	struct declvar *dv = xmalloc(sizeof *dv);
	dv->next = declvar;
	declvar = dv;
	dv->name = name;
	dv->off = off;
	dv->value = value;
}


void
env_init(eval_environ_t env)
{
	int i;

	/* Initialize status and registers */
	env->status = SMFIS_CONTINUE;
	env->tos = 0;
	env->base = 0;
	env->reg = 0;

	env->matches = NULL;
	env->matchsize = 0;
	env->matchcount = 0;
	env->string = NULL;

	/* Initialize catch functions */
	for (i = 0; i < mf_status_count; i++)
		env->catch[i] = 0;

	env_final_gc(env);
}

/* Initialize declared variables */
void
env_decl_vars(eval_environ_t env)
{
	struct declvar *dv;

	for (dv = declvar; dv; dv = dv->next)
		env->var[dv->off] = dv->value;
}

void
env_make_frame(eval_environ_t env)
{
	push(env, (STKVAL) (env->pc + 1));
	push(env, (STKVAL) env->base);
	env->base = env->tos;
}

void
env_leave_frame(eval_environ_t env, int nargs)
{
	env->tos = env->base;
	env->base = (prog_counter_t) pop(env);
	env->pc = (prog_counter_t) pop(env);
	adjust_stack(env, nargs);
}
	
void
env_push_string(eval_environ_t env, char *arg)
{
	push(env, (STKVAL) arg);
}

void
env_push_number(eval_environ_t env, long arg)
{
	push(env, (STKVAL) arg);
}

int
env_set_variable(eval_environ_t env, char *ident, char *value)
{
	char *end;
	struct variable *var = variable_lookup(ident);
	if (!var) {
		mu_error("variable %s is not declared");
		return 1;
	}
	switch (var->type) {
	case dtype_string:
		env->var[var->off] = (STKVAL) strdup(value);
		break;

	case dtype_number:
		env->var[var->off] = (STKVAL) strtol(value, &end, 0);
		if (*end) {
			mu_error("%s is not a number", value);
			return 1;
		}
		break;
		
	default:
		mu_error("type of %s is unknown", ident);
		return 1;
	}
	return 0;
}

int
eval_environment(eval_environ_t env, prog_counter_t start)
{
	if (setjmp(env->x)) 
		return 1;

	for (env->pc = start; ; env->pc++) {
		if (env->pc >= pc)
			runtime_error(env, "pc out of range");
		if (!prog[env->pc])
			break;
		(*(prog[env->pc]))(env);
	}
	return 0;
}

int
env_catch(eval_environ_t env, mf_status status)
{
	prog_counter_t pc = env->catch[status];
	if (pc) {
		/* Reset the exception to avoid recursion. We will never have
		   to restore it again, since the catch handler either returns
		   from the function, in which case the exception mask will be
		   restored, or stops the execution with an action, in which
		   case exceptions are not relevant anymore. */
		env->catch[status] = 0;
		/* Fixup the program counter */
		env->pc = pc - 1;
		/* Generate normal entry frame */
		env_push_number(env, status);
		env_make_frame(env);
		return 0;
	}
	return 1;
}

sfsistat
environment_get_status(eval_environ_t env)
{
	return env->status;
}

const char *
environment_get_null_symbol(eval_environ_t env, struct locus *locus)
{
	*locus = env->locus;
	return env->null_symbol;
}

SMFICTX *
env_get_context(eval_environ_t env)
{
	return env->ctx;
}


/* Capturing support */
int
env_capture_start(eval_environ_t env)
{
	int rc = mu_temp_file_stream_create(&env->stream, NULL);
	if (rc) {
		mu_error("%scannot create temporary stream: %s",
			 mailfromd_msgid(env->ctx), mu_strerror(rc));
		return 1;
	}
	rc = mu_stream_open(env->stream);
	if (rc) {
		mu_error("%scannot open temporary stream: %s",
			 mailfromd_msgid(env->ctx), mu_strerror(rc));
		mu_stream_destroy(&env->stream,
				  mu_stream_get_owner(env->stream));
		return 1;
	}
	return 0;
}
	
int
env_capture_write(eval_environ_t env, const char *buf, size_t size)
{
	if (env->stream) {
		int rc = mu_stream_sequential_write(env->stream, buf, size);
		if (rc) {
			mu_error("%stemporary stream write failed: %s",
				 mailfromd_msgid(env->ctx), mu_strerror(rc));
			mu_stream_close(env->stream);
			mu_stream_destroy(&env->stream,
					  mu_stream_get_owner(env->stream));
		}
		return rc;
	}
	return 1;
}

int
env_capture_write_args(eval_environ_t env, ...)
{
	va_list ap;
	char *arg;
	int rc = 0;

	if (!env->stream)
		return 1;
	va_start(ap, env);
	while (arg = va_arg(ap, char*)) {
		if (rc = env_capture_write(env, arg, strlen(arg)))
			break;
	}
	va_end(ap);
	return rc;
}


struct builtin_priv {  /* Built-in private data structure */
	void *(*init)();
	void (*destroy)(void*);
};

static size_t builtin_priv_size;
static size_t builtin_priv_count;
static struct builtin_priv *bi_priv;

static void
builtin_priv_destroy(void *ptr)
{
	free(ptr);
}

int
builtin_priv_register(void *(*init)(), void (*destroy)(void*))
{
	struct builtin_priv *p;
	int desc;
	
	if (!init)
		return -1;
	if (builtin_priv_count == builtin_priv_size) {
		if (builtin_priv_size == 0)
			builtin_priv_size = 4;
		bi_priv = x2nrealloc(bi_priv, &builtin_priv_size,
				     sizeof bi_priv[0]);
	}
	desc = builtin_priv_count;
	p = bi_priv + builtin_priv_count++;
	p->init = init;
	p->destroy = destroy ? destroy : builtin_priv_destroy;
	return desc;
}

void *
env_get_builtin_priv(eval_environ_t env, int id)
{
	if (id < 0 || id >= builtin_priv_count)
		runtime_error(env,
			      "unknown built-in private data requested (%d)",
			      id);
	if (!env->bi_priv_array) {
		if (builtin_priv_count == 0)
			runtime_error(env,
				      "no built-in private data registered");
		env->bi_priv_array = xcalloc(builtin_priv_count,
					     sizeof env->bi_priv_array[0]);
	}
	if (env->bi_priv_array[id] == NULL) {
		env->bi_priv_array[id] = bi_priv[id].init();
		if (!env->bi_priv_array[id])
			runtime_error(env,
				      "initial allocation for built-in "
				      "private data #%d failed", id);
	}
	return env->bi_priv_array[id];
}

static
env_builtin_priv_destroy(eval_environ_t env)
{
	if (env->bi_priv_array) {
		int i;
		for (i = 0; i < builtin_priv_count; i++) 
			if (env->bi_priv_array[i])
				bi_priv[i].destroy(env->bi_priv_array[i]);
		free(env->bi_priv_array);
	}
}


/* Rudimentary dictionary support */
struct dict_entry {
	char *name;
	char *value;
};

static int
name_comp(const void *item, const void *data)
{
	const struct dict_entry *ent = item;
	const struct dict_entry *p = data;
	return strcmp(ent->name, p->name);
}

void
name_destroy(void *item)
{
	const struct dict_entry *ent = item;
	free(ent->name);
	free(ent->value);
}

void
dict_init(mu_list_t *dict)
{
	mu_list_create(dict);
	mu_list_set_comparator(*dict, name_comp);
	mu_list_set_destroy_item (*dict, name_destroy);
}

char *
dict_install(mu_list_t dict, char *name, char *value)
{
	struct dict_entry ent, *p;

	ent.name = name;

	if (mu_list_locate(dict, &ent, (void**)&p) == 0) {
		free(p->value);
		p->value = strdup(value);
	} else {
		p = malloc(sizeof(*p));
		if (!p)
			abort();
		p->name = strdup(name);
		p->value = strdup(value);
		mu_list_append(dict, p);
	}
	return p->value;
}

void
dict_destroy(mu_list_t *dict)
{
	mu_list_destroy(dict);
}

char *
dict_getsym(void *data, char *str)
{
	struct dict_entry ent, *p;
	char *tmp = NULL;
	char *val = NULL;
	
	if (str[0] == '{') {
		int len = strlen(str);
		if (str[len-1] == '}') {
			tmp = malloc(len-1);
			if (!tmp)
				abort();
			memcpy(tmp, str+1, len-2);
			tmp[len-2] = 0;
			str = tmp;
		}
	}
	
	ent.name = str;
	if (mu_list_locate(data, &ent, (void**)&p) == 0)
		val = p->value;
	free(tmp);
	return val;
}

eval_environ_t
create_environment(SMFICTX *ctx,
		   char *(*getsym)(void *data, char *str),
		   int (*setreply)(void *data, char *code, char *xcode,
				   char *message),
		   void (*setheader)(void *data, struct old_header_node *hdr),
		   void *data)
{
	struct eval_environ *env = calloc(1, sizeof *env);

	if (!env) {
		mu_error("Not enough memory");
		exit(1);
	}

	env->stack = calloc(stack_size, sizeof env->stack[0]);
	if (!env->stack) {
		mu_error("Not enough memory");
		exit(1);
	}

	env->var = calloc(variable_count, sizeof env->var[0]);
	
	env->ctx = ctx;
	env->data = data;
	env->getsym = getsym;
	env->setreply = setreply;
	env->setheader = setheader;
	env->status = SMFIS_CONTINUE;
	/* FIXME:
	   The only register that we initialize here. The rest is initialized
	   in env_init. The top of heap should be retained across calls to
	   handlers, since we store string variables there. This raises stack
	   size requirements. */
	env->toh = stack_size - 1;

	env->bi_priv_array = NULL;
	env_decl_vars(env);
	
	return env;
}

void
destroy_environment(eval_environ_t env)
{
	free(env->stack);
	free(env->var);
	free(env->matches);
	mu_stream_close(env->stream);
	mu_stream_destroy(&env->stream, mu_stream_get_owner(env->stream));
	env_builtin_priv_destroy(env);
	free(env);
}


/* Builtins */

void
builtin_setup ()
{
	struct builtin_module *mod;
	for (mod = builtin_module; mod->name; mod++)
		if (mod->init)
			mod->init();
}




static int
function_counter(void *sym, void *data)
{
	size_t *p = data;
	++*p;
	return 0;
}

struct entry_point {
	int ishandler;
	prog_counter_t pc;
	union {
		enum smtp_state tag;
		const char *name;
	} v;
};

struct enum_data {
	int i;
	struct entry_point *base;
};

static int
function_lister(void *sym, void *data)
{
	struct function *f = sym;
	struct enum_data *d = data;
	d->base[d->i].ishandler = 0;
	d->base[d->i].v.name = f->name;
	d->base[d->i].pc = f->entry;
	d->i++;
	return 0;
}

static int
comp_pc(const void *a, const void *b)
{
	const struct entry_point *ap = a;
	const struct entry_point *bp = b;

	if (ap->pc < bp->pc)
		return -1;
	else if (ap->pc > bp->pc)
		return 1;
	return 0;
}

void
print_code()
{
	enum smtp_state tag;
	struct entry_point *ep; /* Entry points */
	size_t epcount = 0; /* Number of entry points */
	size_t i;
	struct enum_data d;
	prog_counter_t ip;
	
	/* Count all available entry points: */
	for (tag = smtp_state_first; tag < smtp_state_count; tag++)
		if (entry_point[tag])
			epcount++;
	symbol_enumerate(SYM_FUNC, function_counter, &epcount);
	
	ep = xmalloc((epcount+1)*sizeof *ep);

        /* Fill in entry points array */
	i = 0;
	for (tag = smtp_state_first; tag < smtp_state_count; tag++)
		if (entry_point[tag]) {
			ep[i].ishandler = 1;
			ep[i].pc = entry_point[tag];
			ep[i].v.tag = tag;
			i++;
		}

	d.base = ep;
	d.i = i;
	symbol_enumerate(SYM_FUNC, function_lister, &d);

	/* Sort the array */
	qsort(ep, epcount, sizeof ep[0], comp_pc);
	ep[epcount].pc = pc;
	
	/* Actually print the code */

	i = 0;
	while (i < epcount) {
		if (ep[i].ishandler)	
			printf("HANDLER %s\n", state_to_string(ep[i].v.tag));
		else
			printf("FUNCTION %s\n", ep[i].v.name);
		for (ip = ep[i++].pc; ip < ep[i].pc; ip++) { 
			printf("%4lu: ", ip);
			if (prog[ip]) {
				struct optab *op = find_op (prog[ip]);
				if (!op) {
					printf("Unknown function. Aborting dump\n");
					exit (1);
				}
				printf("%s ", op->name);
				if (op->dump)
					op->dump (ip+1, &ip);
				ip += op->length;
			} else
				printf("STOP");
			putchar('\n');
		}
	}
	free(ep);
}


static int
string_counter(void *sym, void *data)
{
	struct variable *var = sym;
	if (var->storage_class == storage_extern
	    && var->type == dtype_string) 
		string_off[string_count++] = var->off;
	return 0;
}

void
post_parse()
{
	string_off = xcalloc(variable_count, sizeof string_off[0]);
	string_count = 0;
	symbol_enumerate(SYM_VARIABLE, string_counter, NULL);
}

static eval_environ_t genv;
static size_t *s_off;
static size_t s_cnt;

static int
s_off_cmp(const void *a, const void *b)
{
	char *pa = genv->var[*(const size_t *) a];
	char *pb = genv->var[*(const size_t *) b];

	if (pa < pb)
		return 1;
	else if (pa > pb)
		return -1;
	return 1;
}

void
env_final_gc(eval_environ_t env)
{
	size_t i;
	char *top = (char*)(env->stack + env->toh);
	char *bot = (char*)(env->stack + stack_size);
	
	genv = env;
	/* Prepare s_off/s_count: remove any variables that are not
	   in heap */
	s_off = xcalloc(string_count, sizeof s_off[0]);
	for (i = 0, s_cnt = 0; i < string_count; i++) {
		char *p = (char*) env->var[string_off[i]];
		if (top < p && p < bot)
			s_off[s_cnt++] = string_off[i];
	}
	qsort(s_off, s_cnt, sizeof s_off[0], s_off_cmp);

	/* Compact the variables */
	env->toh = stack_size - 1;
	for (i = 0; i < s_cnt; i++) {
		char *p = (char*) env->var[s_off[i]];
		size_t len = strlen(p) + 1;
		char *q = (char*)(env->var[s_off[i]] = heap_reserve(env, len));
		memmove(q, p, len);
	}
	free(s_off);
}

